#!/usr/bin/python3
# -*- coding: utf-8 -*-
# --------------------------------------------------------------------
# Copyright © 2014-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

# --------------------------------------------------------------------
# Terminology:
#
# - "update archive": a fake system-image "ubuntu-hash.tar.xz" tar
#   archive.
# --------------------------------------------------------------------

import tarfile
import tempfile
import unittest
import os
import shutil

from unittest.mock import patch

from ubuntucoreupgrader.upgrader import (
    tar_generator,
    Upgrader,
    parse_args,
)

script_name = os.path.basename(__file__)

# file mode to use for creating test directories.
TEST_DIR_MODE = 0o750

CMD_FILE = 'ubuntu_command'


def make_default_options():
    return parse_args([])


def make_tmp_dir(tag=None):
    '''
    Create a temporary directory.
    '''

    if tag:
        prefix = '{}-{}-'.format(script_name, tag)
    else:
        prefix = script_name

    return tempfile.mkdtemp(prefix=prefix)


class UpgradeTestCase(unittest.TestCase):

    def test_tar_generator_unpack_assets(self):
        tempf = tempfile.TemporaryFile()
        with tarfile.TarFile(fileobj=tempf, mode="w") as tar:

            # special top-level file that should not be unpacked
            tar.add(__file__, "removed")

            tar.add(__file__, "system/bin/true")
            tar.add(__file__, "assets/vmlinuz")
            tar.add(__file__, "assets/initrd.img")
            tar.add(__file__, "unreleated/something")
            tar.add(__file__, "hardware.yaml")

            result = [m.name for m in tar_generator(tar, "cache_dir", [], "/")]
            self.assertEqual(result, ["system/bin/true", "assets/vmlinuz",
                                      "assets/initrd.img", "hardware.yaml"])

    def test_tar_generator_system_files_unpack(self):
        tempf = tempfile.TemporaryFile()
        root_dir = make_tmp_dir()
        cache_dir = make_tmp_dir()

        sys_dir = os.path.join(cache_dir, 'system')

        os.makedirs(os.path.join(root_dir, 'dev'))
        os.makedirs(sys_dir)

        with tarfile.TarFile(fileobj=tempf, mode="w") as tar:
            tar.add(__file__, "assets/vmlinuz")
            tar.add(__file__, "assets/initrd.img")

            device_a = '/dev/null'
            path = (os.path.normpath('{}/{}'.format(sys_dir, device_a)))
            touch_file(path)

            self.assertTrue(os.path.exists(os.path.join(root_dir, device_a)))

            # should not be unpacked as already exists
            tar.add(__file__, "system{}".format(device_a))

            device_b = '/dev/does-not-exist'
            self.assertFalse(os.path.exists(os.path.join(root_dir, device_b)))

            # should be unpacked as does not exist
            tar.add(__file__, "system{}".format(device_b))

            expected = ["assets/vmlinuz", "assets/initrd.img",
                        "dev/does-not-exist"]

            expected_results = [os.path.join(root_dir, file)
                                for file in expected]

            result = [m.name for m in
                      tar_generator(tar, cache_dir, [], root_dir)]

            self.assertEqual(result, expected_results)

        shutil.rmtree(root_dir)
        shutil.rmtree(cache_dir)

    def test_tar_generator_system_files_remove_before_unpack(self):
        """
        Test that the upgrader removes certain files just prior to
        overwriting them via the unpack.
        """

        tempf = tempfile.TemporaryFile()
        root_dir = make_tmp_dir()
        cache_dir = make_tmp_dir()
        sys_dir = os.path.join(cache_dir, 'system')

        os.makedirs(sys_dir)

        file = 'a/file'
        dir = 'a/dir'

        with tarfile.TarFile(fileobj=tempf, mode="w") as tar:

            file_path = os.path.normpath('{}/{}'.format(root_dir, file))
            touch_file(file_path)
            self.assertTrue(os.path.exists(file_path))

            dir_path = os.path.normpath('{}/{}'.format(root_dir, dir))
            os.makedirs(dir_path)
            self.assertTrue(os.path.exists(dir_path))

            tar.add(__file__, "system/{}".format(file))

            expected = [file]
            expected_results = [os.path.join(root_dir, f)
                                for f in expected]

            result = [m.name for m in
                      tar_generator(tar, cache_dir, [], root_dir)]

            self.assertEqual(result, expected_results)

            # file should be removed
            self.assertFalse(os.path.exists(file_path))

            # directory should not be removed
            self.assertTrue(os.path.exists(dir_path))

        shutil.rmtree(root_dir)
        shutil.rmtree(cache_dir)


def append_file(path, contents):
    '''
    Append to a regular file (create it doesn't exist).
    '''

    dirname = os.path.dirname(path)
    os.makedirs(dirname, mode=TEST_DIR_MODE, exist_ok=True)

    with open(path, 'a') as fh:
        fh.writelines(contents)

        if not contents.endswith('\n'):
            fh.write('\n')


def create_file(path, contents):
    '''
    Create a regular file.
    '''
    append_file(path, contents)


def touch_file(path):
    '''
    Create an empty file (creating any necessary intermediate
    directories in @path.
    '''
    create_file(path, '')


def make_commands(update_list):
    """
    Convert the specified list of update archives into a list of command
    file commands.
    """
    l = []

    # we don't currently add a mount verb (which would be more
    # realistic) since we can't handle that in the tests.
    # ##l.append('mount system')

    for file in update_list:
        l.append('update {} {}.asc'.format(file, file))

    # we don't currently add an unmount verb (which would be more
    # realistic) since we can't handle that in the tests.
    # ##l.append('unmount system')

    return l


class UpdateTree():
    '''
    Representation of a directory tree that will be converted into an
    update archive.
    '''
    TEST_REMOVED_FILE = 'removed'
    TEST_SYSTEM_DIR = 'system/'

    def __init__(self):

        # Directory tree used to construct the tar file from.
        # Also used to hold the TEST_REMOVED_FILE file.
        self.dir = make_tmp_dir(tag='UpdateTree-tar-source')

        self.removed_file = os.path.join(self.dir, self.TEST_REMOVED_FILE)

        # Directory to place create/modify files into.
        self.system_dir = os.path.join(self.dir, self.TEST_SYSTEM_DIR)

        # Directory used to write the generated tarfile to.
        # This directory should also be used to write the command file
        # to.
        self.tmp_dir = make_tmp_dir(tag='UpdateTree-cache')

    def destroy(self):
        if os.path.exists(self.dir):
            shutil.rmtree(self.dir)

        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def tar_filter(self, member):
        '''
        Function to filter the tarinfo members before creating the
        archive.
        '''
        # members are created with relative paths (no leading slash)
        path = os.sep + member.name

        if member.name == '/.':
            return None

        i = path.find(self.dir)
        assert(i == 0)

        # remove the temporary directory elements
        # (+1 for the os.sep we added above)
        member.name = path[len(self.dir)+1:]

        return member

    def create_archive(self, name):
        '''
        Create an archive with the specified name from the UpdateTree
        object. Also creates a fake signature file alongside the archive
        file since this is currently required by the upgrader (although
        it is not validated).

        :param name: name of tarfile.
        :param name: full path to xz archive to create.
        :return full path to tar file with name @name.
        '''

        tar_path = os.path.join(self.tmp_dir, name)
        tar = tarfile.open(tar_path, 'w:xz')

        # We can't just add recursively since that would attempt to add
        # the parent directory. However, the real update tars don't
        # include that, and attempting to ignore the parent directory
        # results in an empty archive. So, walk the tree and add
        # file-by-file.
        for path, names, files in os.walk(self.dir):
            for file in files:
                full = os.path.join(path, file)
                tar.add(full, recursive=False, filter=self.tar_filter)
            if not files and not names:
                # add (empty) directories
                tar.add(path, recursive=False, filter=self.tar_filter)

        tar.close()

        signature = '{}.asc'.format(tar_path)

        with open(signature, 'w') as fh:
            fh.write('fake signature file')

        return tar_path


class UbuntuCoreUpgraderTestCase(unittest.TestCase):
    '''
    Base class for Upgrader tests.
    '''

    TARFILE = 'update.tar.xz'

    # result of last test run. Hack to deal with fact that even if a
    # test fails, unittest still calls .tearDown() (whomever thought
    # that was a good idea...?)
    currentResult = None

    def setUp(self):
        '''
        Test setup.
        '''
        # Create an object to hold the tree that will be converted into
        # an upgrade archive.
        self.update = UpdateTree()

        # The directory which will have the update archive applied to
        # it.
        self.victim_dir = make_tmp_dir(tag='victim')

    def tearDown(self):
        '''
        Test cleanup.
        '''

        if not self.currentResult.wasSuccessful():
            # Do not clean up - the only sane option if a test fails.
            return

        self.update.destroy()
        self.update = None

        shutil.rmtree(self.victim_dir)
        self.victim_dir = None

    def run(self, result=None):
        self.currentResult = result
        unittest.TestCase.run(self, result)


def mock_get_root_partitions_by_label():
    matches = []

    matches.append(('system-a', '/dev/sda3', '/'))
    matches.append(('system-b', '/dev/sda4', '/writable/cache/system'))

    return matches


def mock_make_mount_private(target):
    pass


class UbuntuCoreUpgraderObectTestCase(UbuntuCoreUpgraderTestCase):

    def mock_get_cache_dir(self):
        '''
        Mock to allow the tests to control the cache directory location.
        '''
        assert(self.cache_dir)
        self.sys_dir = os.path.join(self.cache_dir, 'system')
        os.makedirs(self.sys_dir, exist_ok=True)
        return self.cache_dir

    @patch('ubuntucoreupgrader.upgrader.get_root_partitions_by_label',
           mock_get_root_partitions_by_label)
    @patch('ubuntucoreupgrader.upgrader.make_mount_private',
           mock_make_mount_private)
    def test_create_object(self):
        root_dir = make_tmp_dir()

        file = 'created-regular-file'

        file_path = os.path.join(self.update.system_dir, file)
        create_file(file_path, 'foo bar')

        self.cache_dir = self.update.tmp_dir

        archive = self.update.create_archive(self.TARFILE)
        self.assertTrue(os.path.exists(archive))

        dest = os.path.join(self.cache_dir, os.path.basename(archive))
        touch_file('{}.asc'.format(dest))

        commands = make_commands([self.TARFILE])

        options = make_default_options()

        # XXX: doesn't actually exist, but the option must be set since
        # the upgrader looks for the update archives in the directory
        # where this file is claimed to be.
        options.cmdfile = os.path.join(self.cache_dir, 'ubuntu_command')

        options.root_dir = root_dir

        upgrader = Upgrader(options, commands, [])
        upgrader.get_cache_dir = self.mock_get_cache_dir
        upgrader.run()

        path = os.path.join(root_dir, file)
        self.assertTrue(os.path.exists(path))

        shutil.rmtree(root_dir)

if __name__ == "__main__":
    unittest.main()
